﻿// Goldie: GOLD Engine for D
// Written in the D programming language.

module goldie.exception;

import tango.core.Exception;
import tango.util.Convert;

import semitwist.util.all;
import semitwist.util.compat.all;

import goldie.lexer;
import goldie.parser;
import goldie.token;

//TODO: An easy way for user code to create consistent semantic/etc exceptions
//TODO: Make numErrors a member of this, and display "1 Error(s)" on parse-phase errors.
class ParseException : Exception
{
	Parser parser;
	Lexer lexer;
	
	this(string msg)
	{
		this.parser = null;
		this.lexer = null;
		super(msg);
	}

	this(Parser parser, string msg)
	{
		this.parser = parser;
		this.lexer = parser? parser.lexer : null;
		super(msg);
	}

	protected this(Lexer lexer)
	{
		this.parser = null;
		this.lexer = lexer;
		super(
			reduceTo(
				lexer.errors,
				(string str, LexError error)
				{
					return str ~ error.toString() ~ "\n";
				}
			)~"{} Error(s)".sformat(lexer.errors.length)
		);
	}
}

struct LexError
{
	string file;
	int line;
	int pos;
	string content;
	
	string toString()
	{
		return "{}({}:{}): Syntax Error: '{}'".sformat(file, line+1, pos+1, printableContent);
	}
	
	mixin(getterLazy!(string, "printableContent", `
		return content.stripNonPrintable();
	`));
}

class LexException : ParseException
{
	this(Lexer lexer)
	{
		super(lexer);
	}
}

class InternalException : ParseException
{
	this(string msg)
	{
		super("Internal Goldie Error: " ~ msg);
	}

	this(Parser parser, string msg)
	{
		super(parser, "Internal Goldie Error: " ~ msg);
	}
}

class UnexpectedTokenException : ParseException
{
	this(Parser parser)
	{
		auto token = parser.currTok;
		super(
			parser,
			"{}({}{}): Unexpected {}: '{}'".sformat(
				token.file, token.line+1,
				parser.lexer?
					":{}".sformat((token.srcIndexStart-parser.lexer.lineIndicies[token.line])+1) :
					"",
				token.name, token.toString(TokenToStringMode.Full)
			)
		);
	}
}

class UnexpectedEofException : ParseException
{
	this(Parser parser=null)
	{
		if(parser is null)
			super(parser, "Unexpected EOF");
		else
		{
			super(
				parser,
				"{}({}{}): Unexpected EOF".sformat(
					parser.currTok? parser.currTok.file : parser.filename,
					parser.currTok? to!(string)(parser.currTok.line+1) : "",
					(parser.currTok && parser.lexer)?
						":{}".sformat((parser.currTok.srcIndexStart-parser.lexer.lineIndicies[parser.currTok.line])+1) :
						""
				)
			);
		}
	}
}

class LanguageLoadException : IOException
{
	this(string msg="", string lang="", Exception next=null)
	{
		string fullMsg = "Language '{}'".sformat(lang);
		if(msg != "")
			fullMsg ~= ": " ~ msg;

		super(fullMsg);

		this.next = next;
	}
}

class LanguageNotFoundException : LanguageLoadException
{
	this(string lang, Exception next=null)
	{
		super("Not found", lang);
	}
}
