﻿// Goldie: GOLD Engine for D
// Tools: Dump CGT
// Written in the D programming language.

/++
Dumps a CGT file to human-readable text.

$(WEB http://www.dsource.org/projects/goldie, Goldie Homepage)

Author:
$(WEB www.semitwist.com, Nick Sabalausky)

This has been tested to work with:
  - DMD 1.056 / Tango 0.99.9 / Rebuild 0.76
  - DMD 1.056 / Tango 0.99.9 / xfBuild 0.4
+/

//TODO? Output BOM

module tools.dumpcgt.main;

import tango.io.Stdout;
import tango.io.device.File;

import semitwist.util.all;
import semitwist.util.compat.all;

import goldie.all;

int main(string[] args)
{
	Goldie.langDir = getExecPath()~"../lang/";
		
	if(args.length != 2)
	{
		Stdout("Dump CGT v"~Goldie.verStr).newline;
		Stdout("Copyright (c) 2009-2010 Nick Sabalausky").newline;
		Stdout("See LICENSE.txt for license info").newline;
		Stdout("Site: http://www.dsource.org/projects/goldie").newline;
		Stdout.newline;
		Stdout("Dumps GOLD Parser Builder .cgt files to human-readable text.").newline;
		Stdout.newline;
		Stdout("Usage: dumpcgt languagefile.cgt").newline;
		Stdout("       dumpcgt languagefile.cgt > languagefile.cgt.txt").newline;
		return 1;
	}
	
	string cgtFile = args[1];

	// Load language
	auto lang = Goldie.loadCGT(cgtFile);
	Stdout.formatln("File    : {}", lang.filename);
	Stdout.formatln("Name    : {}", lang.name);
	Stdout.formatln("Version : {}", lang.ver);
	Stdout.formatln("Author  : {}", lang.author);
	Stdout.formatln("About   : {}", lang.about);
	Stdout.formatln("Case Sensitive : {}", lang.caseSensitive);
	Stdout.formatln("");
	
	Stdout.formatln("Start Symbol Index  : {}", lang.startSymbolIndex);
	Stdout.formatln("EOF Symbol Index    : {}", lang.eofSymbolIndex);
	Stdout.formatln("Error Symbol Index  : {}", lang.errorSymbolIndex);

	Stdout.formatln("Symbol Table (Length {}) ->", lang.symbolTable.length);
	foreach(int i, Symbol sym; lang.symbolTable)
	{
		Stdout.formatln(
			"  [{,-4}] : {,-12}: {}",
			i, symbolTypeToString(sym.type), sym.name
		);
	}
	Stdout.formatln("");
	
	Stdout.formatln("Character Set Table (Length {}) ->", lang.charSetTable.length);
	foreach(int i, CharSet charSet; lang.charSetTable)
	{
		dstring charSetStr = charSet.chars;
		Stdout.formatln("  [{,-4}] : {}", i, charSetStr);
	}
	Stdout.formatln("");
	
	Stdout.formatln("Initial DFA State : {}", lang.initialDFAState);
	Stdout.formatln("DFA Table (Length {}) ->", lang.dfaTable.length);
	foreach(int i, DFAState state; lang.dfaTable)
	{
		Stdout.formatln(
			"  [{,-4}] : Accept Symbol #{}: {}, {} Edge{}",
			i, state.acceptSymbolIndex, state.accept, state.edges.length,
			state.edges.length==0? "s" : state.edges.length==1? " ->" : "s ->"
		);
		if(state.edges.length > 0)
		{
			foreach(int iEdge, DFAStateEdge edge; state.edges)
			{
				Stdout.formatln(
					"           [{,-4}] : Char Set #{,-4} : Target DFA State #{,-4}",
					iEdge, edge.charSetIndex, edge.targetDFAStateIndex
				);
			}
		}
	}
	Stdout.formatln("");
	
	Stdout.formatln("Rule Table (Length {}) ->", lang.ruleTable.length);
	foreach(int i, Rule rule; lang.ruleTable)
	{
		string ruleStr = lang.symbolTable[rule.symbolIndex].name ~ " ::=";
		ruleStr =
			reduceTo(rule.subSymbolIndicies, ruleStr,
				(string str, int i)
				{
					return str ~ " " ~ lang.symbolTable[i].name;
				}
			);
		
		Stdout.formatln("  [{,-4}] : {}", i, ruleStr);
	}
	Stdout.formatln("");
	
	Stdout.formatln("Initial LALR State : {}", lang.initialLALRState);
	Stdout.formatln("LALR State Table (Length {}) ->", lang.lalrTable.length);
	foreach(int i, LALRState state; lang.lalrTable)
	{
		Stdout.formatln(
			"  [{,-4}] : {} Action{}", i, state.actions.length,
			state.actions.length==0? "s" : state.actions.length==1? " ->" : "s ->"
		);

		foreach(int iAct, LALRAction action; state.actions)
		{
			string targetStr;
			switch(action.type)
			{
			case LALRAction.Type.Shift:
				targetStr = " to LALR State #{}".sformat(action.target);
				break;
			case LALRAction.Type.Goto:
				targetStr = " LALR State #{}".sformat(action.target);
				break;
			case LALRAction.Type.Reduce:
				targetStr =
					" to {} (Rule #{})".sformat(
						lang.symbolTable[lang.ruleTable[action.target].symbolIndex].name,
						action.target
					);
				break;
			case LALRAction.Type.Accept:
				targetStr = " (Unused Index #{})".sformat(action.target);
				break;
			default:
				throw new Exception("Internal Error: Unexpected action type #{}".sformat(action.type));
			}

			Stdout.formatln(
				"           [{,-4}] : If '{}' (#{}), {}{}",
				iAct, lang.symbolTable[action.symbolId].name, action.symbolId,
				LALRAction.typeToString(action.type), targetStr
			);
		}
	}
	Stdout.formatln("");
	
	return 0;
}
