// Goldie: GOLD Engine for D
// Tools: Generate Documentation
// Written in the D programming language.

module tools.gendocs.cmd;

import tango.io.FilePath;
import tango.io.Path;
import tango.io.Stdout;
import tango.sys.Environment;

import semitwist.cmdlineparser;
import semitwist.util.all;
import semitwist.util.compat.all;

import goldie.all;

class CmdArgs
{
	mixin(getter!(bool, "shouldExit"));

	public this(string[] args)
	{
		init();
		shouldExit = !parse(args);
	}
	
	// Direct command-line args
	public bool help;
	public bool detailhelp;
	public bool saveAST=false;
	public bool quietMode=false;
	public bool trimLinks=false;
	
	public string astDir  = "";
	public string outFile = "index.html";
	public string outDir  = "";
	public string srcFile = "";

	// Generated
	public string srcDir = "";

	// Private ------
	private CmdLineParser cmd;
	
	private void init()
	{
		cmd = new CmdLineParser();
		mixin(defineArg!(cmd, "help",   help,       ArgFlag.Optional, "Displays a help summary and exits" ));
		mixin(defineArg!(cmd, "detail", detailhelp, ArgFlag.Optional, "Displays a detailed help message and exits" ));

		mixin(defineArg!(cmd, "",   srcFile, ArgFlag.Required, "Main template file" ));
		mixin(defineArg!(cmd, "od", outDir,  ArgFlag.Required, "Output directory" ));
		mixin(defineArg!(cmd, "of", outFile, ArgFlag.Optional, "Filename of output pages" ));
		
		mixin(defineArg!(cmd, "trimlink", trimLinks, ArgFlag.Optional, "Trim filename from internal links" ));

		mixin(defineArg!(cmd, "q",    quietMode, ArgFlag.Optional, "Quiet mode" ));
		mixin(defineArg!(cmd, "ast",  saveAST,   ArgFlag.Optional, "Save abstract syntax trees to JSON (for use in JsonViewer)" ));
		mixin(defineArg!(cmd, "astd", astDir,    ArgFlag.Optional, "JSON AST output directory (implies /ast) (default: curr dir)" ));
	}
	
	void outputHeader()
	{
		Stdout.formatln("Goldie GenDocs v{}", Goldie.verStr);
		Stdout("Copyright (c) 2010 Nick Sabalausky").newline;
		Stdout("See LICENSE.txt for license info").newline;
		Stdout("Site: http://www.dsource.org/projects/goldie").newline;
		Stdout.newline;
		Stdout("Generates documentation.").newline;
		Stdout.newline;
	}
	
	// Returns: Should processing proceed? If false, the program should exit.
	private bool parse(string[] args)
	{
		const string sampleUsageMsg = "Sample Usage: gendocs docssrc/Docs.tmpl /od:docs";
		cmd.parse(args);
		if(detailhelp)
		{
			outputHeader();
			Stdout(cmd.errorMsg).newline;
			Stdout(sampleUsageMsg).newline;
			Stdout(cmd.getDetailedUsage());
			return false;
		}
		if(!cmd.success || help)
		{
			outputHeader();
			Stdout(cmd.errorMsg).newline;
			Stdout(sampleUsageMsg).newline;
			Stdout(cmd.getUsage(14));
			return false;
		}

		void ensureTrailingSlash(ref string path)
		{
			// UTF-safe because '/' and '\\' are one code-unit
			if(path[$-1] != '/' && path[$-1] != '\\')
				path ~= "/";
		}
		
		if(astDir != "")
		{
			saveAST = true;
			ensureTrailingSlash(astDir);
		}
			
		if(outDir == "")
			outDir = normalize(getExecPath()~"../../docs/");
		ensureTrailingSlash(outDir);

		if(srcFile == "")
			srcFile = normalize(getExecPath()~"../../docssrc/Docs.tmpl");
		
		auto fpSource = new FilePath(srcFile);
		srcDir  = fpSource.folder;
		srcFile = fpSource.file;
		
		return true;
	}
}
