// Goldie: GOLD Engine for D
// Tools: Parse
// Written in the D programming language.

module tools.parse.cmd;

import tango.io.Stdout;
import tango.io.FilePath;
import tango.io.Path;
import tango.text.Util;

import semitwist.cmdlineparser;
import semitwist.treeout;
import semitwist.util.all;
import semitwist.util.compat.all;

import goldie.all;

class Cmd
{
	mixin(getter!(bool, "shouldExit"));

	public this(string[] args)
	{
		init();
		shouldExit = !parse(args);
	}
	
	// Direct command-line args (may be baked, not guaranteed to be raw)
	public bool   help;
	public bool   detailhelp;
	public string lang;
	public string inputFilename;

	public bool   parseTreePretty=false;
	public string parseTreeOut;
	public string parseTreeType="json";

	public bool   tokensPretty=false;
	public string tokensOut;
	public string tokensType="json";

	public bool  debugOut=false;

	private bool noParseTree=false;
	private bool noTokens=false;

	// Generated
	public TreeFormatter parseFormatter;
	public TreeFormatter tokenFormatter;
	public bool   saveParseTree;
	public bool   saveTokens;
	public string langFile;
	public string langSrc;
	public string langDir;

	// Private ------
	private CmdLineParser cmd;
	
	private void init()
	{
		//TODO: Think about turning defineArg!() into a runtime member func of CmdLineParser
		cmd = new CmdLineParser();
		mixin(defineArg!(cmd, "help",    help,            ArgFlag.Optional, "Displays a help summary and exits" ));
		mixin(defineArg!(cmd, "detail",  detailhelp,      ArgFlag.Optional, "Displays a detailed help message and exits" ));
		mixin(defineArg!(cmd, "",        inputFilename,   ArgFlag.Required, "Input filename" ));
		mixin(defineArg!(cmd, "lang",    lang,            ArgFlag.Optional|ArgFlag.Advanced,   "Language (will load '{lang}.cgt')" ));
		mixin(defineArg!(cmd, "langdir", langDir,         ArgFlag.Optional|ArgFlag.Advanced,   "Directory containing language files" ));
		mixin(defineArg!(cmd, "npt",     noParseTree,     ArgFlag.Optional|ArgFlag.Advanced,   "Don't Save Parse Tree" ));
		mixin(defineArg!(cmd, "ntk",     noTokens,        ArgFlag.Optional|ArgFlag.Advanced,   "Don't Save Tokens" ));
		mixin(defineArg!(cmd, "ptp",     parseTreePretty, ArgFlag.Optional|ArgFlag.Advanced,   `Pretty Format Parse Tree File` ));
		mixin(defineArg!(cmd, "tkp",     tokensPretty,    ArgFlag.Optional|ArgFlag.Advanced,   `Pretty Format Token File` ));
		mixin(defineArg!(cmd, "ptt",     parseTreeType,   ArgFlag.Optional|ArgFlag.Advanced|ArgFlag.ToLower, `Parse Tree File Type ("json" or "xml")` ));
		mixin(defineArg!(cmd, "tkt",     tokensType,      ArgFlag.Optional|ArgFlag.Advanced|ArgFlag.ToLower, `Token File Type ("json" or "xml")` ));
		mixin(defineArg!(cmd, "pto",     parseTreeOut,    ArgFlag.Optional|ArgFlag.Advanced,   "Parse Tree Output Filename" ));
		mixin(defineArg!(cmd, "tko",     tokensOut,       ArgFlag.Optional|ArgFlag.Advanced,   "Token Output Filename" ));
		mixin(defineArg!(cmd, "debug",   debugOut,        ArgFlag.Optional|ArgFlag.Advanced,   "Display Debug Information" ));
		
		mixin(setArgAllowableValues!("ptt", "json", "xml"));
		mixin(setArgAllowableValues!("tkt", "json", "xml"));
	}
	
	void outputHeader()
	{
		Stdout.formatln("Goldie Parse v{}", Goldie.verStr);
		Stdout("Copyright (c) 2009-2010 Nick Sabalausky").newline;
		Stdout("See LICENSE.txt for license info").newline;
		Stdout("Site: http://www.dsource.org/projects/goldie").newline;
		Stdout.newline;
		Stdout("Parses a source file according to a cgt file and saves the result.").newline;
		Stdout.newline;
	}
	
	// Returns: Should processing proceed? If false, the program should exit.
	private bool parse(string[] args)
	{
		const string sampleUsageMsg = "Sample Usage: parse lang/valid_sample2.calc";
		cmd.parse(args);
		if(detailhelp)
		{
			outputHeader();
			Stdout(cmd.errorMsg).newline;
			Stdout(sampleUsageMsg).newline;
			Stdout(cmd.getDetailedUsage());
			return false;
		}
		if(!cmd.success || help)
		{
			outputHeader();
			Stdout(cmd.errorMsg).newline;
			Stdout(sampleUsageMsg).newline;
			Stdout(cmd.getUsage(16));
			return false;
		}
		
		saveParseTree = !noParseTree;
		saveTokens    = !noTokens;

		if(parseTreePretty || parseTreeOut != "")
			saveParseTree = true;
		if(parseTreeOut == "")
			parseTreeOut = inputFilename~".pt."~parseTreeType;
			
		if(tokensPretty || tokensOut != "")
			saveTokens = true;
		if(tokensOut == "")
			tokensOut = inputFilename~".tk."~tokensType;
		
		parseFormatter = saveParseTree? getFormatter(parseTreeType, parseTreePretty) : null;
		tokenFormatter = saveTokens?    getFormatter(   tokensType,    tokensPretty) : null;
		
		if(langDir != "")
		{
			if(langDir[$-1] != '/' && langDir[$-1] != '\\')
				langDir ~= pathSep;
		}
		else
			langDir = Goldie.langDir;
		
		if(lang == "")
		{
			auto indexOfExt = inputFilename.locatePrior('.');
			lang = (indexOfExt<inputFilename.length) ? inputFilename[indexOfExt+1..$] : "";
		}
		
		if(lang.contains('/') || lang.contains('\\'))
		{
			Stdout("Value for '/lang:' must not contain a path.").newline;
			Stdout("To specify a path to the language files, use '/langdir:'.").newline;
			Stdout.newline;
			Stdout(sampleUsageMsg).newline;
			Stdout(cmd.getUsage(16));
			return false;
		}
		
		langFile = langDir~lang~".cgt";
		langSrc  = langDir~lang~".grmp";
		if(!exists(langSrc))
		{
			string langSrcAlt = langDir~lang~".grm";
			if(exists(langSrcAlt))
				langSrc = langSrcAlt;
		}
		
		return true;
	}
	
	private TreeFormatter getFormatter(string type, bool pretty)
	{
		TreeFormatter formatter = 
			(type=="xml"  &&  pretty)? formatterPrettyXML   :
			(type=="xml"  && !pretty)? formatterTrimmedXML  :
			(type=="json" &&  pretty)? formatterPrettyJSON  :
			(type=="json" && !pretty)? formatterTrimmedJSON :
			null;
		return formatter;
	}
}
