// Goldie: GOLD Engine for D
// Tools: Parse
// Written in the D programming language.

/++
Author:
$(WEB www.semitwist.com, Nick Sabalausky)

Parses a source file according to a cgt file, and saves the result
to a JSON file.

XML output is also supported, but the JSON support is much better.

This has been tested to work with:
  - DMD 1.056 / Tango 0.99.9 / Rebuild 0.76
  - DMD 1.056 / Tango 0.99.9 / xfBuild 0.4
+/

//TODO: Why does "Saving Tokens" appear before the lex errors?

module tools.parse.main;

import tango.core.Exception;
import tango.io.Stdout;
import tango.io.device.File;
import tango.text.Util;

import semitwist.treeout;
import semitwist.util.all;
import semitwist.util.compat.all;

import goldie.all;

import tools.parse.cmd;

int main(string[] args)
{
	flushAsserts(); // For unittests
	Goldie.langDir = getExecPath()~"../lang/";
	
	auto cmd = new Cmd(args);
	if(cmd.shouldExit())
		return 1;

	if(cmd.debugOut)
	{
		Stdout.formatln("Input File: {}", cmd.inputFilename);
		Stdout.formatln("Language: {}", cmd.lang);
		Stdout.formatln("Language File: {}", cmd.langFile);
		Stdout.formatln("Language Source: {}", cmd.langSrc);
	}

	auto goldLang = Goldie.loadCGT(cmd.langFile);
	auto parseTree =
		process(
			cmd.inputFilename, goldLang,
			cmd.parseTreeOut, cmd.tokensOut,
			cmd.parseFormatter, cmd.tokenFormatter
		);
	
	if(parseTree is null)
		return 1;
/+
	// Stuff for a future source-processing tool
	Stdout.formatln("------- Getting Ready to Generate AST:");
	Stdout.format("Loading Gold Plus...");
	auto lang = Goldie.loadLanguage("grmp");
	Stdout.formatln("Done!");
	Stdout.format("Loading '{}' Language Source...", cmd.lang);
	auto grmParseTree =
		process(
			cmd.langSrc, lang,
			"", "",
			null, null
		);
	Stdout.formatln("Done!");
	
	//TODO: Associate 'grmParseTree' with 'lang' and apply to parseTree to obtain AST
+/
	return 0;
}

string genTreeStr(Token rootTok, TreeFormatter formatter, string inputFilename, string source)
{
	auto parseTreeRoot = rootTok.toTreeNode();
	parseTreeRoot.addAttribute("file", inputFilename);
	parseTreeRoot.addAttribute("parseTreeMode", "true");
	parseTreeRoot.addAttribute("source", source);
	return parseTreeRoot
	       .format(formatter)
	       .toNativeEOLFromUnix();
}

string genTokenTreeStr(Token[] tokens, Language lang, TreeFormatter formatter, string inputFilename, string source)
{
	return genTreeStr(
		new Token(
			Symbol("Tokens", SymbolType.NonTerminal, -1),
			tokens, lang, -1
		),
		formatter, inputFilename, source
	);
}

// T must be either Token or Token[]
void saveTree(T)(T tokens, Language lang, TreeFormatter formatter, string label,
                 string outFilename, string inputFilename, string source)
{
	if(formatter)
	{
		Stdout.format("Saving {} ({})...", label, outFilename);
		Stdout.flush();
		
		string treeStr;
		static if(is(T:Token))
			treeStr = genTreeStr(tokens, formatter, inputFilename, source);
		else static if(is(T:Token[]))
			treeStr = tokens.genTokenTreeStr(lang, formatter, inputFilename, source);
		else
			static assert(false, "'T tokens' is type '"~T.stringof~"', but must be either type 'Token' or 'Token[]'");
			
		saveFile(outFilename, cast(ubyte[])treeStr);
		Stdout.formatln("Done!");
	}
}

// Returns: Root 'Token' of parse tree, or null if failed to parse.
Token process(string inputFilename, Language lang,
              string parseTreeOut, string tokensOut, 
              TreeFormatter parseFormatter, TreeFormatter tokenFormatter)
{
	Lexer lexer;
	bool lexOk = true;
	try
		lexer = lang.lexFileX(inputFilename);
	catch(ParseException e)
	{
		Stdout(e.msg).newline;
		Stdout.flush();
		lexer = e.lexer;
		lexOk = false;
	}
	auto tokens = lexer.tokens;
	saveTree(tokens, lang, tokenFormatter, "Tokens", tokensOut, inputFilename, lexer.source);
	if(!lexOk)
		return null;

	Parser parser;
	try
		parser = lang.parseTokensX(tokens, inputFilename, lexer);
	catch(ParseException e)
	{
		Stdout(e.msg).newline;
		return null;
	}
	auto parseTree = parser.parseTreeX;
	saveTree(parseTree, lang, parseFormatter, "Parse Tree", parseTreeOut, inputFilename, lexer.source);

	return parseTree;
}

void saveFile(string filename, ubyte[] data)
{
	auto outFile = new File(filename, File.WriteCreate);
	auto bytesWritten = outFile.output.write(data);
	outFile.close();
	
	if(bytesWritten != data.length)
		throw new Exception(
			"Failed to write entire file: {}: Wrote {} out of {} bytes."
				.sformat(filename, bytesWritten, data.length));
}
