﻿// Goldie: GOLD Engine for D
// Tools: StaticLang: Generate Static Goldie Language Interface
// Written in the D programming language.

module tools.staticlang.cmd;

import tango.core.Version;
import tango.io.FilePath;
import tango.io.Stdout;
import tango.text.Util;

import tango.sys.Environment;

import semitwist.cmdlineparser;
import semitwist.util.all;
import semitwist.util.compat.all;

import goldie.all;

string goldiePackageDir;
static this()
{
	goldiePackageDir = getExecPath()~`../src/goldie/`;
}

class CmdArgs
{
	mixin(getter!(bool, "shouldExit"));

	public this(string[] args)
	{
		init();
		shouldExit = !parse(args);
	}
	
	// Direct command-line args (may be baked, not guaranteed to be raw)
	public bool   help;
	public bool   detailhelp;
	public string cgtFile;
	public string outputPackage;
	public string outputDir;

	// Generated
	public string shortPackage; // The last part of the package name
	public string cgtFileOnly;

	// Private ------
	private CmdLineParser cmd;
	
	private void init()
	{
		cmd = new CmdLineParser();
		mixin(defineArg!(cmd, "help",   help,          ArgFlag.Optional, "Displays a help summary and exits" ));
		mixin(defineArg!(cmd, "detail", detailhelp,    ArgFlag.Optional, "Displays a detailed help message and exits" ));
		mixin(defineArg!(cmd, "",       cgtFile,       ArgFlag.Required, "CGT language file to use" ));
		mixin(defineArg!(cmd, "pack",   outputPackage, ArgFlag.Optional, "Name of output package (default: goldie.staticlang.{cgt})" ));
		mixin(defineArg!(cmd, "dir",    outputDir,     ArgFlag.Optional, "Directory of package root (default: current directory)" ));
	}
	
	// Returns: Should processing proceed? If false, the program should exit.
	private bool parse(string[] args)
	{
		string info =
			"Goldie StaticLang v"~Goldie.verStr~"\n"~
			"Copyright (c) 2009-2010 Nick Sabalausky\n"~
			"See LICENSE.txt for license info\n"~
			"Site: http://www.dsource.org/projects/goldie\n"~
			"\n"~
			"Generates static Goldie language interface.";
			
		const string sampleUsageMsg =
			"Sample Usage:\n"~
			"  staticlang langs/lang.cgt\n"~
			"      Creates \"goldie/staticlang/lang/*.d\", each\n"~
			"      containing \"module goldie.staticlang.lang.*;\"\n"~
			"\n"~
			"  staticlang langs/lang.cgt /dir:foo/src /pack:myapp.mylang\n"~
			"      Creates \"foo/src/myapp/mylang/*.d\", each\n"~
			"      containing \"module myapp.mylang.*;\"\n";
		
		cmd.parse(args);
		if(detailhelp)
		{
			Stdout(info).newline.newline;
			Stdout(sampleUsageMsg).newline;
			Stdout(cmd.getDetailedUsage());
			return false;
		}
		if(!cmd.success || help)
		{
			Stdout(info).newline.newline;
			if(!help)
				Stdout(cmd.errorMsg).newline;
			Stdout(sampleUsageMsg).newline;
			Stdout(cmd.getUsage(13));
			return false;
		}
		
		auto fpCgt = new FilePath(cgtFile);
		cgtFileOnly = fpCgt.file;
		
		if(outputPackage == "")
		{
			shortPackage = fpCgt.name;
			outputPackage = "goldie.staticlang."~shortPackage;
		}
		else
		{
			uint i = outputPackage.locatePrior('.');
			if(i == outputPackage.length-1)
				outputPackage = outputPackage[0..$-1];
				
			if(i == outputPackage.length)
				shortPackage = outputPackage;
			else
				shortPackage = outputPackage[i+1..$];
			
			//TODO: Ensure shortPackage (and outputPackage) is a valid ident
		}
		
		if(outputDir == "")
			outputDir = Environment.cwd;

		auto fpGoldie = new FilePath(goldiePackageDir);
		auto fpOut = new FilePath(outputDir);
		fpOut.append(outputPackage.substitute(".", "/"));
		
		if(fpGoldie.native == fpOut.native)
		{
			Stdout.formatln("Error: Output directory is the same as Goldie's package source.");
			Stdout.formatln("This is disallowed in order to prevent overwriting Goldie's source.");
			Stdout.formatln("Please set a different output directory.");
			return 1;
		}
		outputDir = fpOut.native.toString() ~ pathSep;

		return true;
	}
}
